#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import json
import logging
import os
import os.path as osp
import pickle
from functools import cached_property
from json import JSONDecodeError
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import h5py
import numpy as np
from h5py import Dataset as HDFRawDataset
from torch import Tensor
from typing_extensions import override

import torchoutil as to
from torchoutil.extras.hdf.common import (
    _DEFAULTS_RAW_HDF_ATTRIBUTES,
    _DUMPED_JSON_KEYS,
    HDFDatasetAttributes,
    HDFItemType,
)
from torchoutil.extras.numpy.scan_info import numpy_dtype_to_torch_dtype
from torchoutil.nn.functional.indices import get_inverse_perm
from torchoutil.pyoutil.collections import all_eq
from torchoutil.pyoutil.difflib import find_closest_in_list
from torchoutil.pyoutil.inspect import get_current_fn_name
from torchoutil.pyoutil.typing import is_iterable_bytes_or_list, is_iterable_str
from torchoutil.types._typing import ScalarLike
from torchoutil.types.guards import is_scalar_like
from torchoutil.utils.data import DatasetSlicer
from torchoutil.utils.pack.common import _dict_to_tuple
from torchoutil.utils.saving import to_builtin

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=False)

IndexLike = Union[int, Iterable[int], Tensor, slice, None]
ColumnLike = Union[str, Iterable[str], None]
CastMode = Literal[
    "to_torch_or_builtin",
    "to_torch_or_numpy",
    "to_builtin",
    "to_numpy_src",
    "to_torch_src",
    "none",
]
CAST_MODES = (
    "to_torch_or_builtin",
    "to_torch_or_numpy",
    "to_builtin",
    "to_numpy_src",
    "to_torch_src",
    "none",
)

pylog = logging.getLogger(__name__)


class HDFDataset(Generic[T, U], DatasetSlicer[U]):
    def __init__(
        self,
        hdf_fpath: Union[str, Path],
        transform: Optional[Callable[[T], U]] = None,
        keep_padding: Iterable[str] = (),
        return_added_columns: bool = False,
        open_hdf: bool = True,
        cast: CastMode = "none",
        file_kwds: Optional[Dict[str, Any]] = None,
    ) -> None:
        """HDFDataset to read an packed hdf file.

        In HDF, all tensors are padded internally then cropped on-the-fly.
        Several options allows to extract non-padded tensors or return the internal shape of each column.

        Args:
            hdf_fpath: The path to the HDF file.
            transforms: The transform to apply values. default to None.
            keep_padding: Keys to keep padding values. defaults to ().
            return_added_columns: Returns the columns added by pack_to_hdf(.) function.
            open_hdf: If True, open the HDF file at start. defaults to True.
            numpy_to_torch: If True, converts numpy array to PyTorch tensors. defaults to True.
            file_kwds: Options given to h5py file object. defaults to None.
        """
        hdf_fpath = Path(hdf_fpath).resolve().expanduser()
        if not hdf_fpath.is_file():
            names = os.listdir(osp.dirname(hdf_fpath))
            names = [name for name in names if name.endswith(".hdf")]
            names = list(sorted(names))
            names_str = "\n - ".join(names)
            msg = f"Cannot find HDF file in path {hdf_fpath=}."
            if len(names) > 0:
                msg += f" Possible HDF files are:\n - {names_str}"
            raise FileNotFoundError(msg)

        keep_padding = list(keep_padding)
        if file_kwds is None:
            file_kwds = {}

        super().__init__(
            add_indices_support=False,
            add_mask_support=False,
            add_slice_support=False,
        )
        self._hdf_fpath = hdf_fpath
        self._transform = transform
        self._keep_padding = keep_padding
        self._return_added_columns = return_added_columns
        self._cast: CastMode = cast
        self._file_kwds = file_kwds

        self._hdf_file: Any = None

        if open_hdf:
            self.open()

    # Public properties
    @property
    def added_columns(self) -> List[str]:
        """Return the list of columns added by pack_to_hdf function."""
        return self.attrs["added_columns"]

    @property
    def all_columns(self) -> List[str]:
        """The name of all columns of the dataset."""
        return list(self.get_hdf_keys())

    @cached_property
    def attrs(self) -> HDFDatasetAttributes:
        attrs = copy.copy(_DEFAULTS_RAW_HDF_ATTRIBUTES)
        attrs.update(self._hdf_file.attrs)

        for name in _DUMPED_JSON_KEYS:
            try:
                attrs[name] = json.loads(attrs[name])
            except JSONDecodeError:
                pylog.error(f"Cannot load JSON data {attrs[name]=} from {name=}.")

        attrs["added_columns"] = list(attrs["added_columns"])
        attrs["src_np_dtypes"] = {
            k: np.dtype(v) for k, v in attrs["src_np_dtypes"].items()
        }
        return attrs  # type: ignore

    @property
    def column_names(self) -> List[str]:
        """The name of each column of the dataset."""
        column_names = self.all_columns
        column_names = [
            name
            for name in column_names
            if self._return_added_columns or name not in self.added_columns
        ]
        return column_names

    @property
    def info(self) -> Dict[str, Any]:
        """Return the global dataset info."""
        return self.attrs["info"]

    @property
    def item_type(self) -> HDFItemType:
        """Return the global dataset info."""
        return self._hdf_file.attrs.get("item_type", "dict")

    @property
    def keep_padding(self) -> List[str]:
        return self._keep_padding

    @keep_padding.setter
    def keep_padding(self, new_value: Iterable[str]) -> None:
        self._keep_padding = list(new_value)

    @property
    def num_columns(self) -> int:
        return len(self.column_names)

    @property
    def num_rows(self) -> int:
        return len(self)

    @property
    def shape(self) -> Tuple[int, ...]:
        """The shape of the Clotho dataset."""
        return len(self), len(self.column_names)

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    @property
    def user_attrs(self) -> Any:
        return self.attrs["user_attrs"]

    # Private properties to define internal behaviours
    @property
    def _encoding(self) -> str:
        return self.attrs["encoding"]

    @property
    def _shape_suffix(self) -> str:
        """Return the tensor shape suffix in column names."""
        return self.attrs["shape_suffix"]

    @property
    def _src_np_dtypes(self) -> Dict[str, np.dtype]:
        return self.attrs["src_np_dtypes"]

    @cached_property
    def _src_is_unicode(self) -> Dict[str, bool]:
        is_unicode = {name: dt.kind == "U" for name, dt in self._src_np_dtypes.items()}
        return is_unicode

    @property
    def _load_as_complex(self) -> Dict[str, bool]:
        return self.attrs["load_as_complex"]

    # Public methods
    @overload
    def get_item(
        self,
        index: int,
        column: None = None,
    ) -> T:
        ...

    @overload
    def get_item(
        self,
        index: Union[Iterable[int], slice, None],
        column: str,
    ) -> List:
        ...

    @overload
    def get_item(
        self,
        index: Union[Iterable[int], slice, None],
        column: Union[List[str], None] = None,
    ) -> Dict[str, List]:
        ...

    @overload
    def get_item(
        self,
        index: Any,
        column: Any,
        raw: bool = False,
    ) -> Any:
        ...

    @override
    def get_item(
        self,
        index: IndexLike,
        column: ColumnLike = None,
        raw: bool = False,
    ) -> Any:
        if self.is_closed():
            msg = f"Cannot get_raw value with closed HDF file. ({self._hdf_file is not None=} and {bool(self._hdf_file)=})"
            raise RuntimeError(msg)

        if index is None:
            index = slice(None)
        elif is_scalar_like(index):
            index = to.to_item(index)  # type: ignore
        elif isinstance(index, Iterable):
            index = to.to_numpy(index)
        elif isinstance(index, (int, slice)):
            pass
        else:
            raise TypeError(f"Invalid argument type {type(index)=}.")

        if column is None:
            column = self.column_names

        if is_iterable_str(column, accept_str=False):
            result_dict = {
                column_i: self.get_item(index, column_i) for column_i in column
            }
            if self.item_type == "tuple":
                result = _dict_to_tuple(result_dict)
            else:
                result = result_dict

            if (
                isinstance(index, int)
                and self._transform is not None
                and set(column) == set(self.column_names)
            ):
                result = self._transform(result)  # type: ignore
            return result  # type: ignore

        if column not in self.all_columns:
            closest = find_closest_in_list(column, self.all_columns)  # type: ignore
            msg = f"Invalid argument {column=}. (did you mean '{closest}'? Expected one of {tuple(self.all_columns)})"
            raise ValueError(msg)

        if isinstance(index, slice) or (
            isinstance(index, np.ndarray)
            and index.ndim == 1
            and index.dtype.kind in ("b", "i")
        ):
            is_mult = True

        elif isinstance(index, int):
            if not (-len(self) <= index < len(self)):
                msg = f"Invalid argument {index=}. (expected int in range [{-len(self)}, {len(self)-1}])"
                raise IndexError(msg)
            is_mult = False
        else:
            raise TypeError(f"Invalid argument type {type(index)=}.")

        hdf_value = self._get_raw_item(index, column)
        if raw:
            return hdf_value

        if is_mult:
            hdf_values = hdf_value
        else:
            hdf_values = hdf_value[None]
        del hdf_value

        shape_column = f"{column}{self._shape_suffix}"
        must_remove_padding = (
            shape_column in self._hdf_file.keys() and column not in self._keep_padding
        )
        hdf_ds: HDFRawDataset = self._hdf_file[column]
        hdf_dtype: np.dtype = hdf_ds.dtype

        if must_remove_padding:
            shapes = self._get_raw_item(index, shape_column)
            if not is_mult:
                shapes = shapes[None]
            slices_lst = [
                tuple(slice(shape_i) for shape_i in shape) for shape in shapes
            ]
        else:
            slices_lst = [None] * int(hdf_ds.shape[0])

        if (
            self._src_is_unicode.get(column, False)
            or h5py.check_vlen_dtype(hdf_dtype) is str
        ):
            hdf_values = _decode_bytes(hdf_values, encoding=self._encoding)

        if must_remove_padding:
            outputs = []
            for hdf_value, slices in zip(hdf_values, slices_lst):
                # Remove the padding part
                if slices is not None:
                    assert isinstance(hdf_value, np.ndarray) and not isinstance(
                        hdf_value, str
                    )
                    hdf_value = hdf_value[slices]
                hdf_value = self._cast_values(hdf_value, column, hdf_dtype)
                outputs.append(hdf_value)
        else:
            outputs = self._cast_values(hdf_values, column, hdf_dtype)
        del hdf_values

        if not is_mult:
            outputs = outputs[0]
        return outputs

    def close(self, ignore_if_closed: bool = False, remove_file: bool = False) -> None:
        if self.is_closed() and not ignore_if_closed:
            raise RuntimeError("Cannot close the HDF file twice.")

        if not ignore_if_closed:
            self._hdf_file.close()
        if remove_file:
            os.remove(self._hdf_fpath)

        self._hdf_file = None
        self._clear_caches()

    def get_attrs(self) -> HDFDatasetAttributes:
        return self.attrs

    def get_hdf_fpath(self) -> Path:
        return self._hdf_fpath

    def get_hdf_keys(self) -> Tuple[str, ...]:
        if self.is_closed():
            raise RuntimeError("Cannot get keys from a closed HDF file.")
        return tuple(self._hdf_file.keys())

    def get_column_shape(self, column_name: str) -> Tuple[int, ...]:
        if self.is_closed():
            msg = f"Cannot get_column_shape with a closed HDF file. ({self._hdf_file is None=} or {not bool(self._hdf_file)=})"
            raise RuntimeError(msg)
        return tuple(self._hdf_file[column_name].shape)

    def get_columns_shapes(self) -> Dict[str, Tuple[int, ...]]:
        if self.is_closed():
            msg = f"Cannot get_columns_shapes with a closed HDF file. ({self._hdf_file is None=} or {not bool(self._hdf_file)=})"
            raise RuntimeError(msg)

        return {
            column_name: tuple(self._hdf_file[column_name].shape)
            for column_name in self.column_names
        }

    def get_column_dtype(self, column_name: str) -> np.dtype:
        if self.is_closed():
            msg = f"Cannot get dtype with a closed HDF file. ({self._hdf_file is None=} or {not bool(self._hdf_file)=})"
            raise RuntimeError(msg)
        return self._hdf_file[column_name].dtype

    def is_closed(self) -> bool:
        return not self.is_open()

    def is_open(self) -> bool:
        return self._hdf_file is not None and bool(self._hdf_file)

    def open(self, ignore_if_opened: bool = False) -> None:
        if self.is_open():
            if ignore_if_opened:
                return None
            else:
                raise RuntimeError("Cannot open the HDF file twice.")

        self._clear_caches()
        self._hdf_file = h5py.File(self._hdf_fpath, "r", **self._file_kwds)
        self._sanity_check()

    # Magic methods
    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, HDFDataset) and pickle.dumps(self) == pickle.dumps(__o)

    def __enter__(self) -> "HDFDataset":
        return self

    def __exit__(self) -> None:
        if self.is_open():
            self.close()

    @overload
    def __getitem__(self, index: int) -> U:
        ...

    @overload
    def __getitem__(
        self,
        index: Union[Iterable[int], slice, None],
    ) -> Dict[str, list]:
        ...

    @overload
    def __getitem__(self, index: Any) -> Any:
        ...

    def __getitem__(  # type: ignore
        self,
        index: Union[IndexLike, Tuple[IndexLike, ColumnLike]],
    ) -> Any:
        return super().__getitem__(index)  # type: ignore

    def __getstate__(self) -> Dict[str, Any]:
        return {
            "hdf_fpath": self._hdf_fpath,
            "transform": self._transform,
            "keep_padding": self._keep_padding,
            "return_added_columns": self._return_added_columns,
            "cast": self._cast,
            "file_kwds": self._file_kwds,
            "is_open": self.is_open(),
        }

    def __hash__(self) -> int:
        hash_value = 0
        if self.is_open():
            hash_value += self.attrs["global_hash_value"]
        if self._transform is not None:
            hash_value += hash(self._transform)
        hash_value += sum(map(hash, self._keep_padding))
        return hash_value

    def __len__(self) -> int:
        if self.is_closed():
            msg = f"Cannot length of a closed HDF file. ({self._hdf_file is not None=} and {bool(self._hdf_file)=})"
            raise RuntimeError(msg)

        if "length" in self._hdf_file.attrs:
            length = self._hdf_file.attrs["length"]
        elif len(self._hdf_file) > 0:
            hdf_dsets: List[HDFRawDataset] = list(self._hdf_file.values())
            hdf_dsets_lens = [len(ds) for ds in hdf_dsets]
            if not all_eq(hdf_dsets_lens):
                msg = f"Found an different number of lengths in hdf sub-datasets. (found {set(hdf_dsets_lens)})"
                raise ValueError(msg)
            length = hdf_dsets_lens[0]
        else:
            length = 0

        return length

    def __repr__(self) -> str:
        repr_hparams = {"file": osp.basename(self._hdf_fpath), "shape": self.shape}
        repr_ = ", ".join(f"{k}={v}" for k, v in repr_hparams.items())
        return f"{HDFDataset.__name__}({repr_})"

    def __setstate__(self, data: Dict[str, Any]) -> None:
        is_init = hasattr(self, "_hdf_fpath") and hasattr(self, "_hdf_file")
        files_are_different = is_init and self._hdf_fpath != data["hdf_fpath"]
        is_open = is_init and self.is_open()

        if is_init and files_are_different and is_open:
            self.close()

        self._hdf_fpath = data["hdf_fpath"]
        self._transform = data["transform"]
        self._keep_padding = data["keep_padding"]
        self._return_added_columns = data["return_added_columns"]
        self._cast = data["cast"]
        self._file_kwds = data["file_kwds"]

        self._hdf_file = None

        if not is_init or (files_are_different and is_open):
            self.open()

    # Private methods
    def _get_raw_item(
        self,
        index: Union[int, slice, np.ndarray],
        column: str,
    ) -> np.ndarray:
        if isinstance(index, (int, slice)) or (
            isinstance(index, np.ndarray)
            and index.ndim == 1
            and index.dtype.kind == "b"
        ):
            hdf_value: Any = self._hdf_file[column][index]
            if self._load_as_complex.get(column, False):
                hdf_value = to.view_as_complex(hdf_value)
            hdf_value = np.array(hdf_value)

        elif (
            isinstance(index, np.ndarray)
            and index.ndim == 1
            and index.dtype.kind == "i"
        ):
            # Note: slicing with indices required strict sorted list of indices, so we sort + remove duplicates before loading
            local_idxs = np.argsort(index, axis=-1)
            sorted_idxs = index[local_idxs]
            uniq, counts = np.unique(sorted_idxs, return_counts=True)

            hdf_value: Any = self._hdf_file[column][uniq]

            hdf_value = np.repeat(hdf_value, counts, axis=0)
            inv_local_idxs = get_inverse_perm(to.numpy_to_tensor(local_idxs)).numpy()
            hdf_value = hdf_value[inv_local_idxs]

            if self._load_as_complex.get(column, False):
                hdf_value = [to.view_as_complex(value) for value in hdf_value]

        else:
            raise TypeError(f"Invalid argument type {type(index)=}.")

        return hdf_value

    def _sanity_check(self) -> None:
        lens = [dset.shape[0] for dset in self._hdf_file.values()]
        if not all_eq(lens) or lens[0] != len(self):
            msg = (
                f"Incorrect length stored in HDF file. (found {lens=} and {len(self)=})"
            )
            pylog.error(msg)

        if not hasattr(self, "__orig_class__"):
            return None

        t_type = self.__orig_class__.__args__[0]  # type: ignore
        if t_type is not Any and (
            (issubclass(t_type, dict) and self.item_type != "dict")
            or (issubclass(t_type, tuple) and self.item_type != "tuple")
        ):
            msg = f"Invalid HDFDataset typing. (found specified type '{t_type.__name__}' but the internal dataset contains type '{self.item_type}')"
            raise TypeError(msg)

    def _clear_caches(self) -> None:
        if hasattr(self, "attrs"):
            del self.attrs
        if hasattr(self, "_is_unicode"):
            del self._src_is_unicode

    def _cast_values(
        self,
        hdf_values: Union[ScalarLike, np.ndarray, List],
        column: str,
        hdf_dtype: np.dtype,
    ) -> Any:
        if self._cast == "none":
            return hdf_values

        elif self._cast == "to_torch_or_builtin":
            valid = to.shape(hdf_values, return_valid=True).valid
            if valid and hdf_dtype.kind not in ("V", "S", "O"):
                result = to.to_tensor(hdf_values)
            elif isinstance(hdf_values, np.ndarray):
                result = hdf_values.tolist()
            else:
                result = to_builtin(hdf_values)

        elif self._cast == "to_torch_or_numpy":
            valid = to.shape(hdf_values, return_valid=True).valid
            if valid and hdf_dtype.kind not in ("V", "S", "O"):
                result = to.to_tensor(hdf_values)
            else:
                result = np.array(hdf_values)

        elif self._cast == "to_builtin":
            if isinstance(hdf_values, np.ndarray):
                result = hdf_values.tolist()
            else:
                result = to_builtin(hdf_values)

        elif self._cast == "to_numpy_src":
            assert isinstance(hdf_values, np.ndarray), f"{type(hdf_values)=}"
            valid = to.shape(hdf_values, return_valid=True).valid
            src_np_dtypes = self.attrs["src_np_dtypes"]
            target_np_dtype = src_np_dtypes.get(column, hdf_values.dtype)

            if isinstance(hdf_values, np.ndarray):
                result = hdf_values.astype(target_np_dtype)
            elif valid:
                result = np.array(hdf_values, dtype=target_np_dtype)
            else:
                result = hdf_values

        elif self._cast == "to_torch_src":
            assert isinstance(hdf_values, np.ndarray), f"{type(hdf_values)=}"
            valid = to.shape(hdf_values, return_valid=True).valid
            src_np_dtypes = self.attrs["src_np_dtypes"]
            target_np_dtype = src_np_dtypes.get(column, hdf_values.dtype)
            target_pt_dtype = numpy_dtype_to_torch_dtype(target_np_dtype, invalid=None)

            if isinstance(hdf_values, np.ndarray):
                hdf_values_view = hdf_values.view(target_np_dtype)
                result = to.numpy_to_tensor(hdf_values_view)
            elif valid:
                result = to.to_tensor(hdf_values, dtype=target_pt_dtype)
            else:
                result = hdf_values

        else:
            msg = f"Invalid argument {self._cast=}. (expected one of {CAST_MODES})"
            raise ValueError(msg)

        return result


def _decode_bytes(
    encoded: Union[bytes, np.ndarray, Iterable],
    encoding: str,
) -> Union[str, np.ndarray, list]:
    """Decode bytes to str with the specified encoding. Works recursively on list of bytes, list of list of bytes, etc."""
    if isinstance(encoded, (bytes, bytearray)):
        return encoded.decode(encoding=encoding)

    elif isinstance(encoded, np.ndarray):
        if encoded.dtype.kind == "S":
            return np.char.decode(encoded, encoding=encoding)
        elif encoded.dtype.kind == "O" and encoded.ndim > 0:
            return [
                _decode_bytes(encoded_i, encoding=encoding) for encoded_i in encoded
            ]
        else:
            return _decode_bytes(encoded.item(), encoding=encoding)

    elif is_iterable_bytes_or_list(encoded):
        return [_decode_bytes(elt, encoding) for elt in encoded]

    else:
        msg = f"Invalid argument type {type(encoded)} for {get_current_fn_name()}. (expected bytes, bytes ndarray or Iterable)"
        raise TypeError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union

from torchoutil.core.packaging import _NUMPY_AVAILABLE

if not _NUMPY_AVAILABLE:
    from torchoutil.extras.numpy import _numpy_fallback as np  # noqa: F401
    from torchoutil.extras.numpy import _numpy_fallback as numpy

    ACCEPTED_NUMPY_DTYPES = ()

else:
    import numpy  # noqa: F401  # type: ignore
    import numpy as np  # type: ignore

    # Numpy dtypes that can be converted to tensor
    ACCEPTED_NUMPY_DTYPES = (
        np.float64,
        np.float32,
        np.float16,
        np.complex64,
        np.complex128,
        np.int64,
        np.int32,
        np.int16,
        np.int8,
        np.uint8,
        bool,
    )


NumpyNumberLike = Union[np.ndarray, np.number]
NumpyScalarLike = Union[np.ndarray, np.generic]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .classes import (
    BoolLike,
    BuiltinCollection,
    BuiltinNumber,
    BuiltinScalar,
    DataclassInstance,
    EllipsisType,
    NamedTupleInstance,
    NoneType,
    SizedIterable,
    T_BuiltinNumber,
    T_BuiltinScalar,
)
from .guards import (
    is_builtin_number,
    is_builtin_obj,
    is_builtin_scalar,
    is_dataclass_instance,
    is_dict_str,
    is_dict_str_optional_int,
    is_iterable_bool,
    is_iterable_bytes_or_list,
    is_iterable_float,
    is_iterable_int,
    is_iterable_integral,
    is_iterable_iterable_int,
    is_iterable_str,
    is_list_bool,
    is_list_builtin_number,
    is_list_float,
    is_list_int,
    is_list_list_str,
    is_list_number,
    is_list_str,
    is_mapping_str,
    is_namedtuple_instance,
    is_sequence_bool,
    is_sequence_int,
    is_sequence_str,
    is_tuple_int,
    is_tuple_optional_int,
    is_tuple_str,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Literal, Optional, Sequence, Tuple, Union, overload

import torch
from torch import Tensor
from torch.types import Device, Number
from typing_extensions import Never

from torchoutil.core.get import get_device, get_dtype
from torchoutil.pyoutil.typing import BuiltinNumber
from torchoutil.types import (
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    DeviceLike,
    DTypeLike,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)

__all__ = [
    "as_tensor",
    "zeros",
    "ones",
    "full",
    "empty",
    "rand",
]


# Empty lists
@overload
def as_tensor(
    data: Sequence[Never],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Never]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[Never]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor3D:
    ...


# bool
@overload
def as_tensor(
    data: bool,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor0D:
    ...


@overload
def as_tensor(
    data: Sequence[bool],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[bool]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[bool]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor3D:
    ...


# int
@overload
def as_tensor(
    data: int,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor0D:
    ...


@overload
def as_tensor(
    data: Sequence[int],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[int]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[int]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor3D:
    ...


# float
@overload
def as_tensor(
    data: float,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor0D:
    ...


@overload
def as_tensor(
    data: Sequence[float],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[float]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[float]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor3D:
    ...


# complex
@overload
def as_tensor(
    data: complex,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor0D:
    ...


@overload
def as_tensor(
    data: Sequence[complex],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[complex]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[complex]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor3D:
    ...


# BuiltinNumber
@overload
def as_tensor(
    data: BuiltinNumber,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor0D:
    ...


@overload
def as_tensor(
    data: Sequence[BuiltinNumber],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor1D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor2D:
    ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor3D:
    ...


def as_tensor(
    data: Any,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.as_tensor(data, dtype=dtype, device=device)


@overload
def zeros(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D:
    ...


@overload
def zeros(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def zeros(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def zeros(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


@overload
def zeros(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def zeros(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def zeros(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


def zeros(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.zeros(
        *data,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


@overload
def ones(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D:
    ...


@overload
def ones(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def ones(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def ones(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


@overload
def ones(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def ones(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def ones(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


def ones(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.ones(
        *data,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


@overload
def empty(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D:
    ...


@overload
def empty(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def empty(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def empty(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


@overload
def empty(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def empty(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def empty(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


def empty(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.empty(
        *data,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


@overload
def rand(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D:
    ...


@overload
def rand(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def rand(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def rand(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


@overload
def rand(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def rand(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def rand(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


def rand(
    *data: Any,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.rand(
        *data,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


# bool
@overload
def full(
    size: Sequence[Never],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor0D:
    ...


@overload
def full(
    size: Tuple[int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor1D:
    ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor2D:
    ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor3D:
    ...


# int
@overload
def full(
    size: Sequence[Never],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor0D:
    ...


@overload
def full(
    size: Tuple[int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor1D:
    ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor2D:
    ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor3D:
    ...


# float
@overload
def full(
    size: Sequence[Never],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor0D:
    ...


@overload
def full(
    size: Tuple[int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor1D:
    ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor2D:
    ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor3D:
    ...


# complex
@overload
def full(
    size: Sequence[Never],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor0D:
    ...


@overload
def full(
    size: Tuple[int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor1D:
    ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor2D:
    ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor3D:
    ...


# BuiltinNumber
@overload
def full(
    size: Sequence[Never],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D:
    ...


@overload
def full(
    size: Tuple[int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D:
    ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D:
    ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D:
    ...


def full(
    size: Sequence[int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = get_dtype(dtype)
    device = get_device(device)
    return torch.full(
        size,
        fill_value=fill_value,
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


@overload
def arange(
    end: int,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: Optional[Device] = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D:
    ...


@overload
def arange(
    start: int,
    end: int,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: Optional[Device] = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D:
    ...


@overload
def arange(
    start: int,
    end: int,
    step: int,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: Optional[Device] = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D:
    ...


def arange(
    start: Number,
    end: Number,
    step: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Optional[torch.dtype] = None,
    device: Optional[Device] = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor1D:
    return torch.arange(
        start,
        end,
        step,
        out=out,
        dtype=dtype,
        device=device,
        requires_grad=requires_grad,
        pin_memory=pin_memory,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
from typing import Any, Callable, Dict
from typing import Generator as PythonGenerator
from typing import (
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Sized,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from torch import Tensor, nn
from typing_extensions import TypeGuard

from torchoutil.core.get import get_device
from torchoutil.extras.numpy import (
    ACCEPTED_NUMPY_DTYPES,
    np,
    numpy_all_eq,
    numpy_all_ne,
    numpy_is_complex,
    numpy_is_floating_point,
    numpy_view_as_complex,
    numpy_view_as_real,
)
from torchoutil.pyoutil.collections import all_eq as builtin_all_eq
from torchoutil.pyoutil.collections import all_ne as builtin_all_ne
from torchoutil.pyoutil.collections import is_sorted as builtin_is_sorted
from torchoutil.pyoutil.collections import prod as builtin_prod
from torchoutil.pyoutil.collections import unzip
from torchoutil.pyoutil.functools import identity
from torchoutil.pyoutil.typing import (
    BuiltinScalar,
    SizedIterable,
    T_BuiltinNumber,
    is_builtin_number,
    is_builtin_scalar,
)
from torchoutil.types._typing import (
    ComplexFloatingTensor,
    FloatingTensor,
    LongTensor,
    ScalarLike,
    T_TensorLike,
)
from torchoutil.types.guards import is_list_tensor, is_scalar_like, is_tuple_tensor
from torchoutil.utils import return_types

T = TypeVar("T")
U = TypeVar("U")


def count_parameters(
    model: nn.Module,
    *,
    recurse: bool = True,
    only_trainable: bool = False,
    buffers: bool = False,
) -> int:
    """Returns the number of parameters in a module."""
    params = (
        param
        for param in model.parameters(recurse)
        if not only_trainable or param.requires_grad
    )

    if buffers:
        params = itertools.chain(params, (buffer for buffer in model.buffers(recurse)))

    num_params = sum(param.numel() for param in params)
    return num_params


def find(
    x: Tensor,
    value: Any,
    default: Union[None, Tensor, int, float] = None,
    dim: int = -1,
) -> LongTensor:
    """Return the index of the first occurrence of value in a tensor."""
    if x.ndim == 0:
        msg = f"Function find does not supports 0-d tensors. (found {x.ndim=} == 0)"
        raise ValueError(msg)

    mask = x.eq(value)
    contains = mask.any(dim=dim)
    indices = mask.long().argmax(dim=dim)

    if default is None:
        if not contains.all():
            raise RuntimeError(f"Cannot find {value=} in tensor.")
        return indices  # type: ignore
    else:
        output = torch.where(contains, indices, default)
        return output  # type: ignore


@overload
def move_to_rec(
    x: Mapping[T, U],
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Dict[T, U]:
    ...


@overload
def move_to_rec(
    x: T,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> T:
    ...


def move_to_rec(
    x: Any,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Any:
    """Move all modules and tensors recursively to a specific dtype or device."""
    if "device" in kwargs:
        kwargs["device"] = get_device(kwargs["device"])

    if isinstance(x, (str, float, int, bool, complex)):
        return x
    elif isinstance(x, (Tensor, nn.Module)):
        if predicate is None or predicate(x):
            return x.to(**kwargs)
        else:
            return x
    elif isinstance(x, Mapping):
        return {k: move_to_rec(v, predicate=predicate, **kwargs) for k, v in x.items()}
    elif isinstance(x, Iterable):
        generator = (move_to_rec(xi, predicate=predicate, **kwargs) for xi in x)
        if isinstance(x, PythonGenerator):
            return generator
        elif isinstance(x, tuple):
            return tuple(generator)
        else:
            return list(generator)
    else:
        return x


def can_be_stacked(
    tensors: Union[List[Any], Tuple[Any, ...]],
) -> TypeGuard[Union[List[Tensor], Tuple[Tensor, ...]]]:
    """Returns True if inputs can be passed to `torch.stack` function, i.e. contanis a list or tuple of tensors with the same shape.

    Alias of :func:`~torchoutil.nn.functional.others.is_stackable`.
    """
    return is_stackable(tensors)


def is_stackable(
    tensors: Union[List[Any], Tuple[Any, ...]],
) -> TypeGuard[Union[List[Tensor], Tuple[Tensor, ...]]]:
    """Returns True if inputs can be passed to `torch.stack` function, i.e. contanis a list or tuple of tensors with the same shape."""
    if not is_list_tensor(tensors) and not is_tuple_tensor(tensors):
        return False
    if len(tensors) == 0:
        return False
    shape0 = tensors[0].shape
    result = all(tensor.shape == shape0 for tensor in tensors[1:])
    return result


def can_be_converted_to_tensor(x: Any) -> bool:
    """Returns True if inputs can be passed to `torch.as_tensor` function.

    Alias of :func:`~torchoutil.nn.functional.others.is_convertible_to_tensor`.

    This function returns False for heterogeneous inputs like `[[], 1]`, but this kind of value can be accepted by `torch.as_tensor`.
    """
    return is_convertible_to_tensor(x)


def is_convertible_to_tensor(x: Any) -> bool:
    """Returns True if inputs can be passed to `torch.as_tensor` function.

    This function returns False for heterogeneous inputs like `[[], 1]`, but this kind of value can be accepted by `torch.as_tensor`.
    """
    if isinstance(x, Tensor):
        return True
    else:
        return __can_be_converted_to_tensor_nested(x)


@overload
def ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    return_valid: Literal[False] = False,
) -> int:
    ...


@overload
def ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    return_valid: Literal[True],
) -> return_types.ndim:
    ...


def ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    return_valid: bool = False,
    use_first_for_list_tuple: bool = False,
) -> Union[int, return_types.ndim]:
    """Scan first argument to return its number of dimension(s). Works recursively with Tensors, numpy arrays and builtins types instances.

    Note: Sets and dicts are considered as scalars with a shape equal to 0.

    Args:
        x: Input value to scan.
        return_valid: If True, returns a tuple containing a boolean indicator if the data has an homogeneous ndim instead of raising a ValueError.
        use_first_for_list_tuple: If True, use first value to determine ndim for list and tuple argument. Otherwise it will scan each value in argument to determine its shape. defaults to False.

    Raises:
        ValueError if input has an heterogeneous number of dimensions.
        TypeError if input has an unsupported type.
    """

    def _impl(
        x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    ) -> Tuple[bool, int]:
        if is_scalar_like(x):
            return True, 0
        elif isinstance(x, (Tensor, np.ndarray, np.generic)):
            return True, x.ndim
        elif isinstance(x, (set, frozenset, dict)):
            return True, 0
        elif isinstance(x, (list, tuple)):
            valids_and_ndims = unzip(_impl(xi) for xi in x)  # type: ignore
            if len(valids_and_ndims) == 0:
                return True, 1

            valids, ndims = valids_and_ndims
            if (use_first_for_list_tuple and valids[0]) or (
                all(valids) and builtin_all_eq(ndims)
            ):
                return True, ndims[0] + 1
            else:
                return False, -1
        else:
            raise TypeError(f"Invalid argument type {type(x)}.")

    valid, ndim = _impl(x)
    if return_valid:
        return return_types.ndim(valid, ndim)
    elif valid:
        return ndim
    else:
        msg = f"Invalid argument {x}. (cannot compute ndim for heterogeneous data)"
        raise ValueError(msg)


@overload
def shape(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    return_valid: Literal[False] = False,
) -> T:
    ...


@overload
def shape(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    return_valid: Literal[True],
) -> return_types.shape[T]:
    ...


def shape(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    return_valid: bool = False,
    use_first_for_list_tuple: bool = False,
) -> Union[T, return_types.shape[T]]:
    """Scan first argument to return its shape. Works recursively with Tensors, numpy arrays and builtins types instances.

    Note: Sets and dicts are considered as scalars with a shape equal to ().

    Args:
        x: Input value to scan.
        output_type: Output shape type. defaults to identity, which returns a tuple of ints.
        return_valid: If True, returns a tuple containing a boolean indicator if the data has an homogeneous shape instead of raising a ValueError.
        use_first_for_list_tuple: If True, use first value to determine ndim for list and tuple argument. Otherwise it will scan each value in argument to determine its shape. defaults to False.

    Raises:
        ValueError: if input has an heterogeneous shape.
        TypeError: if input has an unsupported type.
    """

    def _impl(
        x: Union[ScalarLike, Tensor, np.ndarray, Iterable]
    ) -> Tuple[bool, Tuple[int, ...]]:
        if is_scalar_like(x):
            return True, ()
        elif isinstance(x, (Tensor, np.ndarray, np.generic)):
            return True, tuple(x.shape)
        elif isinstance(x, (set, frozenset, dict)):
            return True, ()
        elif isinstance(x, (list, tuple)):
            valids_and_shapes = unzip(_impl(xi) for xi in x)  # type: ignore
            if len(valids_and_shapes) == 0:
                return True, (0,)

            valids, shapes = valids_and_shapes
            if (use_first_for_list_tuple and valids[0]) or (
                all(valids) and builtin_all_eq(shapes)
            ):
                return True, (len(shapes),) + shapes[0]
            else:
                return False, ()
        else:
            raise TypeError(f"Invalid argument type {type(x)}.")

    valid, shape = _impl(x)
    if return_valid:
        shape = output_type(shape)
        return return_types.shape(valid, shape)
    elif valid:
        shape = output_type(shape)
        return shape
    else:
        msg = f"Invalid argument {x}. (cannot compute shape for heterogeneous data)"
        raise ValueError(msg)


def to_item(x: Union[ScalarLike, Tensor, np.ndarray, SizedIterable]) -> BuiltinScalar:
    """Convert scalar value to built-in type."""
    if is_builtin_scalar(x, strict=True):
        return x
    elif isinstance(x, (Tensor, np.ndarray, np.generic)) and nelement(x) == 1:
        return x.item()
    elif isinstance(x, SizedIterable) and len(x) == 1:
        return to_item(next(iter(x)))
    else:
        msg = f"Invalid argument type {type(x)=}. (expected scalar-like object)"
        raise TypeError(msg)


def ranks(x: Tensor, dim: int = -1, descending: bool = False) -> Tensor:
    """Get the ranks of each value in range [0, x.shape[dim][."""
    return x.argsort(dim, descending).argsort(dim)


def nelement(x: Union[ScalarLike, Tensor, np.ndarray, Iterable]) -> int:
    """Returns the number of elements in Tensor-like object."""
    if isinstance(x, Tensor):
        return x.nelement()
    elif isinstance(x, (np.ndarray, np.generic)):
        return x.size
    else:
        return builtin_prod(shape(x))


def __can_be_converted_to_tensor_list_tuple(x: Union[List, Tuple]) -> bool:
    if len(x) == 0:
        return True

    valid_items = all(__can_be_converted_to_tensor_nested(xi) for xi in x)
    if not valid_items:
        return False

    # If all values are scalars-like items
    if all(
        (not isinstance(xi, Sized) or (isinstance(xi, Tensor) and xi.ndim == 0))
        for xi in x
    ):
        return True

    # If all values are sized items with same size
    elif all(isinstance(xi, Sized) for xi in x):
        len0 = len(x[0])
        return all(len(xi) == len0 for xi in x[1:])

    else:
        return False


def __can_be_converted_to_tensor_nested(
    x: Any,
) -> bool:
    if is_builtin_number(x):
        return True
    elif isinstance(x, Tensor) and x.ndim == 0:
        return True
    elif isinstance(x, (np.ndarray, np.generic)) and x.dtype in ACCEPTED_NUMPY_DTYPES:
        return True
    elif isinstance(x, (List, Tuple)):
        return __can_be_converted_to_tensor_list_tuple(x)
    else:
        return False


@overload
def view_as_real(x: Tensor) -> Tensor:
    ...


@overload
def view_as_real(x: np.ndarray) -> np.ndarray:
    ...


@overload
def view_as_real(x: complex) -> Tuple[float, float]:
    ...


def view_as_real(
    x: Union[Tensor, np.ndarray, complex]
) -> Union[Tensor, np.ndarray, Tuple[float, float]]:
    """Convert complex-valued input to floating-point object."""
    if isinstance(x, Tensor):
        return torch.view_as_real(x)

    elif isinstance(x, np.ndarray):
        return numpy_view_as_real(x)
    else:
        return x.real, x.imag


@overload
def view_as_complex(x: Tensor) -> ComplexFloatingTensor:
    ...


@overload
def view_as_complex(x: np.ndarray) -> np.ndarray:
    ...


@overload
def view_as_complex(x: Tuple[float, float]) -> complex:
    ...


def view_as_complex(
    x: Union[Tensor, np.ndarray, Tuple[float, float]]
) -> Union[ComplexFloatingTensor, np.ndarray, complex]:
    """Convert floating-point input to complex-valued object."""
    if isinstance(x, Tensor):
        return torch.view_as_complex(x)  # type: ignore
    elif isinstance(x, np.ndarray):
        return numpy_view_as_complex(x)
    elif (
        isinstance(x, Sequence)
        and len(x) == 2
        and isinstance(x[0], float)
        and isinstance(x[1], float)
    ):
        return x[0] + x[1] * 1j
    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


@overload
def prod(
    x: Tensor,
    *,
    dim: Optional[int] = None,
    start: Any = 1,
) -> Tensor:
    ...


@overload
def prod(
    x: Iterable[T_BuiltinNumber],
    *,
    dim: Any = None,
    start: T_BuiltinNumber = 1,
) -> T_BuiltinNumber:
    ...


def prod(
    x: Union[Tensor, Iterable[T_BuiltinNumber]],
    *,
    dim: Optional[int] = None,
    start: T_BuiltinNumber = 1,
) -> Union[Tensor, T_BuiltinNumber]:
    """Returns the product of all elements in input."""
    if isinstance(x, Tensor):
        return torch.prod(x, dim=dim)
    elif isinstance(x, np.ndarray):
        return np.prod(x, axis=dim)
    elif isinstance(x, Iterable):
        if dim is not None:
            msg = f"Invalid argument {dim=}. (expected None with {type(x)=})"
            raise ValueError(msg)
        return builtin_prod(x, start=start)  # type: ignore
    else:
        msg = (
            f"Invalid argument type {type(x)=}. (expected Tensor, ndarray or Iterable)"
        )
        raise TypeError(msg)


def is_floating_point(x: Any) -> TypeGuard[Union[FloatingTensor, np.ndarray, float]]:
    """Returns True if object is a/contains floating-point object(s)."""
    if isinstance(x, Tensor):
        return x.is_floating_point()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_is_floating_point(x)
    else:
        return isinstance(x, float)


def is_complex(x: Any) -> TypeGuard[Union[ComplexFloatingTensor, np.ndarray, complex]]:
    """Returns True if object is a/contains complex-valued object(s)."""
    if isinstance(x, Tensor):
        return x.is_complex()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_is_complex(x)
    else:
        return isinstance(x, complex)


def is_sorted(
    x: Union[Tensor, np.ndarray, Iterable],
    *,
    reverse: bool = False,
    strict: bool = False,
) -> bool:
    """Returns True if the sequence is sorted."""
    if isinstance(x, (Tensor, np.ndarray)):
        if x.ndim != 1:
            msg = f"Invalid number of dims in argument {x.ndim=}. (expected 1)"
            raise ValueError(msg)

        if not reverse and not strict:
            result = (x[:-1] <= x[1:]).all().item()
        elif not reverse and strict:
            result = (x[:-1] < x[1:]).all().item()
        elif reverse and not strict:
            result = (x[:-1] >= x[1:]).all().item()
        else:  # reverse and strict
            result = (x[:-1] > x[1:]).all().item()
        return result  # type: ignore

    elif isinstance(x, Iterable):
        return builtin_is_sorted(x, reverse=reverse, strict=strict)

    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


@overload
def all_eq(
    x: Union[Tensor, np.ndarray, ScalarLike, Iterable],
    dim: None = None,
) -> bool:
    ...


@overload
def all_eq(
    x: Union[T_TensorLike],
    dim: int,
) -> T_TensorLike:
    ...


def all_eq(
    x: Union[T_TensorLike, ScalarLike, Iterable],
    dim: Union[int, None] = None,
) -> Union[bool, T_TensorLike]:
    """Check if all elements are equal in a tensor, ndarray, iterable or scalar object."""
    if isinstance(x, Tensor):
        if dim is None:
            if x.ndim == 0 or x.nelement() == 0:
                return True
            x = x.reshape(-1)
            return (x[0] == x[1:]).all().item()
        else:
            slices = [slice(None) for _ in range(x.ndim)]
            slices[dim] = 0
            slices.insert(dim + 1, None)
            return (x == x[slices]).all(dim)

    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_all_eq(x, dim=dim)

    elif dim is not None:
        raise ValueError(f"Invalid argument {dim=} with {type(x)=}.")

    elif is_scalar_like(x):
        return True

    elif isinstance(x, Iterable):
        return builtin_all_eq(x)

    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


def all_ne(x: Union[Tensor, np.ndarray, ScalarLike, Iterable]) -> bool:
    """Check if all elements are NOT equal in a tensor, ndarray, iterable or scalar object."""
    if isinstance(x, Tensor):
        return len(torch.unique(x)) == x.nelement()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_all_ne(x)
    elif is_scalar_like(x):
        return True
    elif isinstance(x, Iterable):
        return builtin_all_ne(x)
    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


def average_power(
    x: T_TensorLike,
    dim: Union[int, Tuple[int, ...], None] = -1,
) -> T_TensorLike:
    """Compute average power of a signal along a specified dim/axis."""
    return (abs(x) ** 2).mean(dim)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    overload,
)

from torchoutil.core.packaging import _PANDAS_AVAILABLE
from torchoutil.pyoutil.csv import load_csv as load_csv_base
from torchoutil.pyoutil.csv import to_csv as to_csv_base
from torchoutil.utils.saving.common import to_builtin

if _PANDAS_AVAILABLE:
    import pandas as pd

CSVBackend = Literal["csv", "pandas"]
CSV_BACKENDS = ("csv", "pandas")
ORIENTS = ("list", "dict")


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["dict"],
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "csv",
    **csv_reader_kwds,
) -> Dict[str, List[Any]]:
    ...


@overload
def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "csv",
    **csv_reader_kwds,
) -> List[Dict[str, Any]]:
    ...


def load_csv(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list", "dict"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "csv",
    **csv_reader_kwds,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]]]:
    if backend == "csv":
        return load_csv_base(
            fpath,
            orient=orient,
            header=header,
            comment_start=comment_start,
            strip_content=strip_content,
            **csv_reader_kwds,
        )

    elif backend == "pandas":
        return _load_csv_with_pandas(
            fpath,
            orient=orient,
            header=header,
            comment_start=comment_start,
            strip_content=strip_content,
            **csv_reader_kwds,
        )

    else:
        msg = f"Invalid argument {backend=}. (expected one of {CSV_BACKENDS})"
        raise ValueError(msg)


def _load_csv_with_pandas(
    fpath: Union[str, Path],
    /,
    *,
    orient: Literal["list", "dict"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    **csv_reader_kwds,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]]]:
    backend = "pandas"

    if not _PANDAS_AVAILABLE:
        msg = f"Invalid argument {backend=} without pandas installed."
        raise ValueError(msg)

    if strip_content:
        msg = f"Invalid argument {strip_content=} with {backend=}."
        raise ValueError(msg)

    if comment_start is not None:
        msg = f"Invalid argument {comment_start=} with {backend=}."
        raise ValueError(msg)

    if len(csv_reader_kwds) > 0:
        msg = f"Invalid arguments {csv_reader_kwds=} with {backend=}."
        raise ValueError(msg)

    df = pd.read_csv(fpath)  # type: ignore

    if orient == "list":
        return df.to_dict("records")  # type: ignore
    elif orient == "dict":
        return df.to_dict("list")  # type: ignore
    else:
        msg = f"Invalid argument {orient=}. (expected one of {ORIENTS})"
        raise ValueError(msg)


def to_csv(
    data: Union[Iterable[Mapping[str, Any]], Mapping[str, Iterable[Any]]],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    header: bool = True,
    **csv_writer_kwds,
) -> str:
    """Dump content to csv format."""
    if to_builtins:
        data = to_builtin(data)

    return to_csv_base(
        data,
        fpath,
        overwrite=overwrite,
        make_parents=make_parents,
        header=header,
        **csv_writer_kwds,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.pad import cat_padded_batch, pad_and_stack_rec


class TestPad(TestCase):
    def test_pad_sequence_rec_example_1(self) -> None:
        sequence = [[1, 2], [3], [], [4, 5]]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.as_tensor([[1, 2], [3, 0], [0, 0], [4, 5]])
        assert output.ndim == output_expected.ndim
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_example_2(self) -> None:
        sequence_invalid = [[1, 2, 3], 3]
        self.assertRaises(ValueError, pad_and_stack_rec, sequence_invalid, 0)

    def test_pad_sequence_rec_example_3(self) -> None:
        sequence = [[[1], [2, 3]], [[4, 5, 6], [7], []]]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.as_tensor(
            [[[1, 0, 0], [2, 3, 0], [0, 0, 0]], [[4, 5, 6], [7, 0, 0], [0, 0, 0]]]
        )
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_example_4(self) -> None:
        sequence = [torch.zeros(10, 2), torch.zeros(5, 2)]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.zeros(2, 10, 2)
        assert output.ndim == output_expected.ndim
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_example_5(self) -> None:
        sequence = [torch.zeros(3, 5), torch.zeros(3, 10)]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.zeros(2, 3, 10)
        assert output.ndim == output_expected.ndim
        self.assertTrue(
            output.eq(output_expected).all(), f"{output}\n{output_expected}"
        )

    def test_pad_sequence_rec_example_6(self) -> None:
        sequence = [torch.zeros(3, 2, 5), torch.zeros(3, 6, 5)]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.zeros(2, 3, 6, 5)
        assert output.ndim == output_expected.ndim
        self.assertTrue(
            output.eq(output_expected).all(), f"{output}\n{output_expected}"
        )

    def test_pad_sequence_rec_example_7(self) -> None:
        sequence = [torch.zeros(4, 2, 3, 10), torch.zeros(4, 5, 5, 10)]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.zeros(2, 4, 5, 5, 10)
        assert output.ndim == output_expected.ndim
        self.assertTrue(
            output.eq(output_expected).all(), f"{output}\n{output_expected}"
        )

    def test_pad_sequence_rec_example_8(self) -> None:
        sequence = [torch.zeros(5, 10, 2), torch.zeros(2, 10, 2), torch.zeros(2, 10, 2)]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.zeros(3, 5, 10, 2)
        assert output.ndim == output_expected.ndim
        self.assertTrue(
            output.eq(output_expected).all(), f"{output}\n{output_expected}"
        )

    def test_pad_sequence_rec_example_9(self) -> None:
        sequence = [[torch.ones(20, 1), torch.ones(8, 2)], [torch.ones(1, 10)]]
        output = pad_and_stack_rec(sequence, 1)
        output_expected = torch.ones(2, 2, 20, 10)
        assert output.ndim == output_expected.ndim
        self.assertTrue(
            output.eq(output_expected).all(), f"{output}\n{output_expected}"
        )

    def test_pad_sequence_rec_limit_1(self) -> None:
        sequence = []
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.as_tensor([])
        assert output.ndim == output_expected.ndim
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_limit_2(self) -> None:
        sequence = [[]]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.as_tensor([[]])
        assert output.ndim == output_expected.ndim
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_limit_3(self) -> None:
        sequence = [[[], []], [[], []], [[], []]]
        output = pad_and_stack_rec(sequence, 0)
        output_expected = torch.as_tensor([[[], []], [[], []], [[], []]])
        assert output.ndim == output_expected.ndim
        assert output.eq(output_expected).all()

    def test_pad_sequence_rec_limit_4(self) -> None:
        sequence = [[], [[]]]
        self.assertRaises(ValueError, pad_and_stack_rec, sequence, 0)


class TestCatPaddedBatch(TestCase):
    def test_example_1(self) -> None:
        # Cas 2D
        x1 = torch.as_tensor(
            [
                [1, 1, 0, 0],
                [2, 2, 2, 0],
                [3, 0, 0, 0],
            ]
        )
        x2 = torch.as_tensor(
            [
                [4, 4, 4, 4, 4],
                [5, 5, 0, 0, 0],
                [6, 6, 0, 0, 0],
            ]
        )
        x1_lens = torch.as_tensor([2, 3, 1])
        x2_lens = torch.as_tensor([5, 2, 2])
        seq_dim = -1
        batch_dim = 0

        expected = torch.as_tensor(
            [
                [1, 1, 4, 4, 4, 4, 4],
                [2, 2, 2, 5, 5, 0, 0],
                [3, 6, 6, 0, 0, 0, 0],
            ]
        )
        expected_lens = torch.as_tensor([7, 5, 3])
        x12, x12_lens = cat_padded_batch(x1, x1_lens, x2, x2_lens, seq_dim, batch_dim)

        assert expected.shape == x12.shape
        assert expected_lens.shape == x12_lens.shape
        assert torch.equal(expected, x12)
        assert torch.equal(expected_lens, x12_lens)

    def test_example_2(self) -> None:
        # Cas 3D
        x1 = torch.as_tensor(
            [
                [[1, 11], [1, 11], [0, 00], [0, 00]],
                [[2, 22], [2, 22], [2, 22], [0, 00]],
                [[3, 33], [0, 00], [0, 00], [0, 00]],
            ]
        )
        x2 = torch.as_tensor(
            [
                [[4, 44], [4, 44], [4, 44], [4, 44], [4, 44]],
                [[5, 55], [5, 55], [0, 00], [0, 00], [0, 00]],
                [[6, 66], [6, 66], [0, 00], [0, 00], [0, 00]],
            ]
        )
        x1_lens = torch.as_tensor([2, 3, 1])
        x2_lens = torch.as_tensor([5, 2, 2])
        seq_dim = -2
        batch_dim = 0

        expected = torch.as_tensor(
            [
                [[1, 11], [1, 11], [4, 44], [4, 44], [4, 44], [4, 44], [4, 44]],
                [[2, 22], [2, 22], [2, 22], [5, 55], [5, 55], [0, 00], [0, 00]],
                [[3, 33], [6, 66], [6, 66], [0, 00], [0, 00], [0, 00], [0, 00]],
            ]
        )
        expected_lens = torch.as_tensor([7, 5, 3])
        x12, x12_lens = cat_padded_batch(x1, x1_lens, x2, x2_lens, seq_dim, batch_dim)

        assert expected.shape == x12.shape
        assert expected_lens.shape == x12_lens.shape
        assert torch.equal(expected, x12)
        assert torch.equal(expected_lens, x12_lens)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import operator
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import TypeIs

from torchoutil.pyoutil.typing import T_BuiltinScalar, is_builtin_scalar, is_mapping_str

K = TypeVar("K", covariant=True)
T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)
V = TypeVar("V", covariant=True)
W = TypeVar("W", covariant=True)
X = TypeVar("X", covariant=True)

KeyMode = Literal["intersect", "same", "union"]
KEY_MODES = ("same", "intersect", "union")


@overload
def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: Literal["intersect", "same"],
    default_val: Any = None,
    *,
    default_val_fn: Any = None,
) -> Dict[K, List[V]]:
    ...


@overload
def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: Literal["union"],
    default_val: Any = None,
    *,
    default_val_fn: Callable[[K], X],
) -> Dict[K, List[Union[V, X]]]:
    ...


@overload
def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: Literal["union"],
    default_val: W = None,
    *,
    default_val_fn: None = None,
) -> Dict[K, List[Union[V, W]]]:
    ...


def list_dict_to_dict_list(
    lst: Sequence[Mapping[K, V]],
    key_mode: KeyMode,
    default_val: W = None,
    default_val_fn: Optional[Callable[[K], X]] = None,
) -> Dict[K, List[Union[V, W]]]:
    """Convert list of dicts to dict of lists.

    Args:
        lst: The list of dict to merge.
        key_mode: Can be "same" or "intersect".
            If "same", all the dictionaries must contains the same keys otherwise a ValueError will be raised.
            If "intersect", only the intersection of all keys will be used in output.
            If "union", the output dict will contains the union of all keys, and the missing value will use the argument default_val.
        default_val: Default value of an element when key_mode is "union". defaults to None.
    """
    if len(lst) <= 0:
        return {}

    keys = set(lst[0].keys())
    if key_mode == "same":
        invalids = [list(item.keys()) for item in lst[1:] if keys != set(item.keys())]
        if len(invalids) > 0:
            msg = f"Invalid dict keys for conversion from List[dict] to Dict[list]. (with {key_mode=}, {keys=} and {invalids=})"
            raise ValueError(msg)

    elif key_mode == "intersect":
        keys = intersect_lists([item.keys() for item in lst])

    elif key_mode == "union":
        keys = union_lists(item.keys() for item in lst)

    else:
        msg = f"Invalid argument key_mode={key_mode}. (expected one of {KEY_MODES})"
        raise ValueError(msg)

    result = {
        key: [
            item.get(
                key, default_val_fn(key) if default_val_fn is not None else default_val
            )
            for item in lst
        ]
        for key in keys
    }
    return result


@overload
def dict_list_to_list_dict(
    dic: Mapping[T, Iterable[U]],
    key_mode: Literal["same", "intersect"],
    default_val: Any = None,
) -> List[Dict[T, U]]:
    ...


@overload
def dict_list_to_list_dict(
    dic: Mapping[T, Iterable[U]],
    key_mode: Literal["union"] = "union",
    default_val: W = None,
) -> List[Dict[T, Union[U, W]]]:
    ...


def dict_list_to_list_dict(
    dic: Mapping[T, Iterable[U]],
    key_mode: KeyMode = "union",
    default_val: W = None,
) -> List[Dict[T, Union[U, W]]]:
    """Convert dict of lists with same sizes to list of dicts.

    Example 1
    ----------
    ```
    >>> dic = {"a": [1, 2], "b": [3, 4]}
    >>> dict_list_to_list_dict(dic)
    ... [{"a": 1, "b": 3}, {"a": 2, "b": 4}]
    ```

    Example 2
    ----------
    ```
    >>> dic = {"a": [1, 2, 3], "b": [4], "c": [5, 6]}
    >>> dict_list_to_list_dict(dic, key_mode="union", default=-1)
    ... [{"a": 1, "b": 4, "c": 5}, {"a": 2, "b": -1, "c": 6}, {"a": 3, "b": -1, "c": -1}]
    ```
    """
    if len(dic) == 0:
        return []

    dic = {k: list(v) if not isinstance(v, Sequence) else v for k, v in dic.items()}
    lengths = [len(seq) for seq in dic.values()]

    if key_mode == "same":
        if not all_eq(lengths):
            msg = f"Invalid sequences for batch. (found different lengths in sub-lists: {set(lengths)})"
            raise ValueError(msg)
        length = lengths[0]

    elif key_mode == "intersect":
        length = min(lengths)

    elif key_mode == "union":
        length = max(lengths)

    else:
        msg = f"Invalid argument key_mode={key_mode}. (expected one of {KEY_MODES})"
        raise ValueError(msg)

    result = [
        {k: (v[i] if i < len(v) else default_val) for k, v in dic.items()}
        for i in range(length)
    ]
    return result


def intersect_lists(lst_of_lst: Sequence[Iterable[T]]) -> List[T]:
    """Performs intersection of elements in lists (like set intersection), but keep their original order."""
    if len(lst_of_lst) <= 0:
        return []
    out = list(dict.fromkeys(lst_of_lst[0]))
    for lst_i in lst_of_lst[1:]:
        out = [name for name in out if name in lst_i]
        if len(out) == 0:
            break
    return out


def union_lists(lst_of_lst: Iterable[Iterable[T]]) -> List[T]:
    """Performs union of elements in lists (like set union), but keep their original order."""
    out = {}
    for lst_i in lst_of_lst:
        out.update(dict.fromkeys(lst_i))
    out = list(out)
    return out


def dump_dict(
    dic: Mapping[str, T],
    /,
    join: str = ", ",
    fmt: str = "{key}={value}",
    ignore_lst: Iterable[T] = (),
) -> str:
    """Custom dictionary of scalars to string function to customize representation.

    Example 1:
    ----------
    ```
    >>> d = {"a": 1, "b": 2}
    >>> dump_dict(d)
    ... 'a=1, b=2'
    ```
    """
    ignore_lst = dict.fromkeys(ignore_lst)
    result = join.join(
        fmt.format(key=key, value=value)
        for key, value in dic.items()
        if value not in ignore_lst
    )
    return result


@overload
def find(
    x: T,
    include: Iterable[V],
    *,
    match_fn: Callable[[V, T], bool] = operator.eq,
    order: Literal["right"] = "right",
    default: U = -1,
) -> Union[int, U]:
    ...


@overload
def find(
    x: T,
    include: Iterable[V],
    *,
    match_fn: Callable[[T, V], bool] = operator.eq,
    order: Literal["left"],
    default: U = -1,
) -> Union[int, U]:
    ...


def find(
    x: T,
    include: Iterable[T],
    *,
    match_fn: Callable[[T, T], bool] = operator.eq,
    order: Literal["left", "right"] = "right",
    default: U = -1,
) -> Union[int, U]:
    if order == "right":
        pass
    elif order == "left":

        def revert(f):
            def reverted_f(a, b):
                return f(b, a)

            return reverted_f

        match_fn = revert(match_fn)
    else:
        ORDER_VALUES = ("left", "right")
        raise ValueError(f"Invalid argument {order=}. (expected one of {ORDER_VALUES})")

    for i, include_i in enumerate(include):
        if match_fn(include_i, x):
            return i
    return default


def contained(
    x: T,
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
    *,
    match_fn: Callable[[T, T], bool] = operator.eq,
    order: Literal["left", "right"] = "right",
) -> bool:
    """Returns True if name in include set and not in exclude set."""
    if (
        include is not None
        and find(x, include, match_fn=match_fn, order=order, default=-1) == -1
    ):
        return False

    if (
        exclude is not None
        and find(x, exclude, match_fn=match_fn, order=order, default=-1) != -1
    ):
        return False

    return True


def filter_iterable(
    it: Iterable[T],
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
    *,
    match_fn: Callable[[T, T], bool] = operator.eq,
    order: Literal["left", "right"] = "right",
) -> List[T]:
    return [
        item
        for item in it
        if contained(
            item,
            include=include,
            exclude=exclude,
            match_fn=match_fn,
            order=order,
        )
    ]


def all_eq(it: Iterable[T], eq_fn: Optional[Callable[[T, T], bool]] = None) -> bool:
    """Returns true if all elements in iterable are equal.

    Note: This function returns True for iterable that contains 0 or 1 element.
    """
    it = list(it)
    first = it[0]
    if eq_fn is None:
        return all(first == elt for elt in it)
    else:
        return all(eq_fn(first, elt) for elt in it)


def all_ne(
    it: Iterable[T],
    ne_fn: Optional[Callable[[T, T], bool]] = None,
    use_set: bool = False,
) -> bool:
    """Returns true if all elements in iterable are differents.

    Note: This function returns True for iterable that contains 0 or 1 element.
    """
    if use_set and ne_fn is not None:
        raise ValueError(f"Cannot use arguments {use_set=} with {ne_fn=}.")

    it = list(it)
    if use_set:
        return len(it) == len(set(it))
    elif ne_fn is None:
        return all(
            it[i] != it[j] for i in range(len(it)) for j in range(i + 1, len(it))
        )
    else:
        return all(
            ne_fn(it[i], it[j]) for i in range(len(it)) for j in range(i + 1, len(it))
        )


def flat_dict_of_dict(
    nested_dic: Mapping[str, Any],
    *,
    sep: str = ".",
    flat_iterables: bool = False,
    overwrite: bool = True,
) -> Dict[str, Any]:
    """Flat a nested dictionary.

    Example 1
    ---------
    ```
    >>> dic = {
    ...     "a": 1,
    ...     "b": {
    ...         "a": 2,
    ...         "b": 10,
    ...     },
    ... }
    >>> flat_dict_of_dict(dic)
    ... {"a": 1, "b.a": 2, "b.b": 10}
    ```

    Example 2
    ---------
    ```
    >>> dic = {"a": ["hello", "world"], "b": 3}
    >>> flat_dict_of_dict(dic, flat_iterables=True)
    ... {"a.0": "hello", "a.1": "world", "b": 3}
    ```

    Args:
        nested_dic: Nested mapping containing sub-mappings or iterables.
        sep: Separators between keys.
        flat_iterables: If True, flat iterable and use index as key.
        overwrite: If True, overwrite duplicated keys in output. Otherwise duplicated keys will raises a ValueError.
    """

    def _impl(nested_dic: Mapping[str, Any]) -> Dict[str, Any]:
        output = {}
        for k, v in nested_dic.items():
            if is_mapping_str(v):
                v = _impl(v)
                v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
                output.update(v)

            elif flat_iterables and isinstance(v, Iterable) and not isinstance(v, str):
                v = {f"{i}": vi for i, vi in enumerate(v)}
                v = _impl(v)
                v = {f"{k}{sep}{kv}": vv for kv, vv in v.items()}
                output.update(v)

            elif overwrite or k not in output:
                output[k] = v

            else:
                msg = f"Ambiguous flatten dict with key '{k}'. (with value '{v}')"
                raise ValueError(msg)
        return output

    return _impl(nested_dic)


def unflat_dict_of_dict(dic: Mapping[str, Any], *, sep: str = ".") -> Dict[str, Any]:
    """Unflat a dictionary.

    Example 1
    ----------
    ```
    >>> dic = {
        "a.a": 1,
        "b.a": 2,
        "b.b": 3,
        "c": 4,
    }
    >>> unflat_dict_of_dict(dic)
    ... {"a": {"a": 1}, "b": {"a": 2, "b": 3}, "c": 4}
    ```
    """
    output = {}
    for k, v in dic.items():
        if sep not in k:
            output[k] = v
        else:
            idx = k.index(sep)
            k, kk = k[:idx], k[idx + 1 :]
            if k not in output:
                output[k] = {}
            elif not isinstance(output[k], Mapping):
                msg = f"Invalid dict argument. (found keys {k} and {k}{sep}{kk})"
                raise ValueError(msg)

            output[k][kk] = v

    output = {
        k: (unflat_dict_of_dict(v) if isinstance(v, Mapping) else v)
        for k, v in output.items()
    }
    return output


@overload
def flat_list_of_list(
    lst: Iterable[Sequence[T]],
    return_sizes: Literal[True] = True,
) -> Tuple[List[T], List[int]]:
    ...


@overload
def flat_list_of_list(
    lst: Iterable[Sequence[T]],
    return_sizes: Literal[False],
) -> List[T]:
    ...


def flat_list_of_list(
    lst: Iterable[Sequence[T]],
    return_sizes: bool = True,
) -> Union[Tuple[List[T], List[int]], List[T]]:
    """Return a flat version of the input list of sublists with each sublist size."""
    flatten_lst = [elt for sublst in lst for elt in sublst]
    sizes = [len(sents) for sents in lst]

    if return_sizes:
        return flatten_lst, sizes
    else:
        return flatten_lst


def unflat_list_of_list(
    flatten_lst: Sequence[T],
    sizes: Iterable[int],
) -> List[List[T]]:
    """Unflat a list to a list of sublists of given sizes."""
    lst = []
    start = 0
    stop = 0
    for count in sizes:
        stop += count
        lst.append(flatten_lst[start:stop])
        start = stop
    return lst


@overload
def unzip(lst: Iterable[Tuple[()]]) -> Tuple[()]:
    ...


@overload
def unzip(lst: Iterable[Tuple[T]]) -> Tuple[List[T]]:
    ...


@overload
def unzip(lst: Iterable[Tuple[T, U]]) -> Tuple[List[T], List[U]]:
    ...


@overload
def unzip(lst: Iterable[Tuple[T, U, V]]) -> Tuple[List[T], List[U], List[V]]:
    ...


@overload
def unzip(
    lst: Iterable[Tuple[T, U, V, W]]
) -> Tuple[List[T], List[U], List[V], List[W]]:
    ...


@overload
def unzip(
    lst: Iterable[Tuple[T, U, V, W, X]]
) -> Tuple[List[T], List[U], List[V], List[W], List[X]]:
    ...


def unzip(lst):
    """Invert zip() function.

    .. code-block:: python
        :caption:  Example

        >>> lst1 = [1, 2, 3, 4]
        >>> lst2 = [5, 6, 7, 8]
        >>> lst_zipped = list(zip(lst1, lst2))
        >>> lst_zipped
        ... [(1, 5), (2, 6), (3, 7), (4, 8)]
        >>> unzip(lst_zipped)
        ... [1, 2, 3, 4], [5, 6, 7, 8]
    """
    return tuple(map(list, zip(*lst)))


def prod(x: Iterable[T], /, start: T = 1) -> T:
    result = copy.copy(start)
    for xi in x:
        result = result * xi  # type: ignore
    return result


def sorted_dict(
    x: Mapping[K, V],
    /,
    *,
    key: Optional[Callable[[K], Any]] = None,
    reverse: bool = False,
) -> Dict[K, V]:
    return {k: x[k] for k in sorted(x.keys(), key=key, reverse=reverse)}  # type: ignore


@overload
def flatten(
    x: T_BuiltinScalar,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]:
    ...


@overload
def flatten(
    x: Iterable[T_BuiltinScalar],
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]:
    ...


@overload
def flatten(
    x: Any,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
    is_scalar_fn: Callable[[Any], TypeIs[T]] = is_builtin_scalar,
) -> List[Any]:
    ...


def flatten(
    x: Any,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
    is_scalar_fn: Callable[[Any], TypeIs[T]] = is_builtin_scalar,
) -> List[Any]:
    if end_dim is None:
        end_dim = sys.maxsize
    if start_dim < 0:
        raise ValueError(f"Invalid argument {start_dim=}. (expected positive integer)")
    if end_dim < 0:
        raise ValueError(f"Invalid argument {end_dim=}. (expected positive integer)")
    if start_dim > end_dim:
        msg = f"Invalid arguments {start_dim=} and {end_dim=}. (expected start_dim <= end_dim)"
        raise ValueError(msg)

    def flatten_impl(x: Any, start_dim: int, end_dim: int) -> List[Any]:
        if is_scalar_fn(x):
            return [x]
        elif isinstance(x, Iterable):
            if start_dim > 0:
                return [flatten_impl(xi, start_dim - 1, end_dim - 1) for xi in x]
            elif end_dim > 0:
                return [
                    xij
                    for xi in x
                    for xij in flatten_impl(xi, start_dim - 1, end_dim - 1)
                ]
            else:
                return list(x)
        else:
            raise TypeError(f"Invalid argument type {type(x)=}.")

    return flatten_impl(x, start_dim, end_dim)


def recursive_generator(x: Any) -> Generator[Tuple[Any, int, int], None, None]:
    def recursive_generator_impl(
        x: Any,
        i: int,
        deep: int,
    ) -> Generator[Tuple[Any, int, int], None, None]:
        if isinstance(x, Iterable):
            for j, xj in enumerate(x):
                if xj == x:
                    yield xj, i, deep
                    return
                else:
                    yield from recursive_generator_impl(xj, j, deep + 1)
        else:
            yield x, i, deep
        return

    return recursive_generator_impl(x, 0, 0)


def is_sorted(
    x: Iterable[Any],
    *,
    reverse: bool = False,
    strict: bool = False,
) -> bool:
    it = iter(x)
    try:
        prev = next(it)
    except StopIteration:
        return True

    for xi in it:
        if not reverse and prev > xi:
            return False
        if reverse and prev < xi:
            return False
        if strict and prev == xi:
            return False
        prev = xi
    return True


def union_dicts(dicts: Iterable[Mapping[K, V]]) -> Dict[K, V]:
    result = {}
    for dic in dicts:
        result.update(dic)
    return result


def argmin(x: Iterable) -> int:
    min_index, _max_value = min(enumerate(x), key=lambda t: t[1])
    return min_index


def argmax(x: Iterable) -> int:
    max_index, _max_value = max(enumerate(x), key=lambda t: t[1])
    return max_index



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import tempfile
import unittest
from pathlib import Path
from unittest import TestCase

from torch import Tensor

from torchoutil.hub.registry import RegistryHub


class TestRegistryHub(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

    def test_cnext_register(self) -> None:
        register = RegistryHub(
            infos={
                "cnext_bl_70": {
                    "architecture": "ConvNeXt",
                    "url": "https://zenodo.org/record/8020843/files/convnext_tiny_465mAP_BL_AC_70kit.pth?download=1",
                    "hash_value": "0688ae503f5893be0b6b71cb92f8b428",
                    "hash_type": "md5",
                    "fname": "convnext_tiny_465mAP_BL_AC_70kit.pth",
                    "state_dict_key": "model",
                },
            },
            register_root=self.tmpdir,
        )

        model_name = "cnext_bl_70"
        register.download_file(model_name, force=False)
        state_dict = register.load_state_dict(
            model_name,
            offline=True,
            load_kwds=dict(map_location="cpu"),
        )

        assert isinstance(state_dict, dict)
        assert all(isinstance(k, str) for k in state_dict.keys())
        assert all(isinstance(v, Tensor) for v in state_dict.values())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any

from torchoutil.pyoutil.pickle import load_pickle  # noqa: F401
from torchoutil.pyoutil.pickle import to_pickle as to_pickle_base
from torchoutil.utils.saving.common import to_builtin


def to_pickle(obj: Any, fpath: Path, *, to_builtins: bool = False) -> bytes:
    if to_builtins:
        obj = to_builtin(obj)
    return to_pickle_base(obj, fpath)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import tempfile
import time
import unittest
from pathlib import Path
from unittest import TestCase

import torch
from torch import Generator
from torch.utils.data.dataset import Subset
from torchvision.datasets import CIFAR10

from torchoutil.extras.numpy import np
from torchoutil.nn import ESequential, IndexToOnehot, ToList, ToNumpy
from torchoutil.utils.pack.pack import pack_dataset, pack_dataset_to_columns


class TestPackCIFAR10(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        dataset = CIFAR10(
            tmpdir,
            train=False,
            transform=ToNumpy(),
            target_transform=ESequential(IndexToOnehot(10), ToList()),
            download=True,
        )
        cls.dataset = dataset
        cls.tmpdir = tmpdir

    def test_cifar10_pack_per_item(self) -> None:
        cls = self.__class__
        dataset = cls.dataset
        tmpdir = cls.tmpdir

        seed = int(torch.randint(0, 10000, ()).item())
        generator = Generator().manual_seed(seed)
        dataset = Subset(
            dataset,
            torch.randint(
                0,
                len(dataset),
                (max(len(dataset) // 10, 1),),
                generator=generator,
            ).tolist(),
        )

        path = tmpdir.joinpath("test_cifar10")
        pkl_dataset = pack_dataset(
            dataset,
            path,
            exists="overwrite",
        )

        assert len(dataset) == len(pkl_dataset)

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = pkl_dataset[idx]

        assert len(dataset) == len(pkl_dataset)
        assert label0 == label1, f"{label0=}, {label1=}"
        assert np.equal(image0, image1).all()

    def test_cifar10_pack_per_batch(self) -> None:
        cls = self.__class__
        dataset = cls.dataset
        tmpdir = cls.tmpdir

        seed = int(torch.randint(0, 10000, ()).item())
        generator = Generator().manual_seed(seed)
        dataset = Subset(
            dataset,
            torch.randint(
                0,
                len(dataset),
                (max(len(dataset) // 10, 1),),
                generator=generator,
            ).tolist(),
        )

        path = tmpdir.joinpath("test_cifar10_batch")
        pkl_dataset = pack_dataset(
            dataset,
            path,
            exists="overwrite",
            content_mode="batch",
        )

        assert len(dataset) == len(pkl_dataset)

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = pkl_dataset[idx]

        assert len(dataset) == len(pkl_dataset)
        assert label0 == label1, f"{label0=}, {label1=}"
        assert np.equal(image0, image1).all()

    def test_cifar10_pack_subdir(self) -> None:
        cls = self.__class__
        dataset = cls.dataset
        tmpdir = cls.tmpdir

        seed = int(torch.randint(0, 10000, ()).item())
        generator = Generator().manual_seed(seed)
        dataset = Subset(
            dataset,
            torch.randint(
                0,
                len(dataset),
                (max(len(dataset) // 50, 1),),
                generator=generator,
            ).tolist(),
        )

        path = tmpdir.joinpath("test_cifar10_subdir")
        pkl_dataset = pack_dataset(
            dataset,
            path,
            exists="overwrite",
            content_mode="item",
            subdir_size=100,
        )

        assert len(dataset) == len(pkl_dataset)

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = pkl_dataset[idx]

        assert len(dataset) == len(pkl_dataset)
        assert label0 == label1, f"{label0=}, {label1=}"
        assert np.equal(image0, image1).all()

    def test_cifar10_pack_columns(self) -> None:
        cls = self.__class__
        dataset = cls.dataset
        tmpdir = cls.tmpdir

        path = tmpdir.joinpath("test_cifar10_columns")
        packed = pack_dataset_to_columns(
            dataset,
            path,
            exists="overwrite",
            save_fn="numpy",
            ds_kwds=dict(load_fn="numpy"),
        )

        assert len(dataset) == len(packed)
        for i in range(len(dataset)):
            sample_i, label_i = dataset[i]
            psample_i, plabel_i = packed[i]

            assert np.all(sample_i == psample_i)
            assert np.all(label_i == plabel_i)


class TestPackSpeechCommands(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = Path(os.getenv("TORCHOUTIL_TMPDIR", tempfile.gettempdir())).joinpath(
            "torchoutil_tests"
        )
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

    def test_example_1(self) -> None:
        from torch import nn
        from torchaudio.datasets import SPEECHCOMMANDS
        from torchaudio.transforms import Spectrogram

        from torchoutil.utils.pack import pack_dataset

        speech_commands_root = self.tmpdir.joinpath("speech_commands")
        packed_root = self.tmpdir.joinpath("packed_speech_commands")

        os.makedirs(speech_commands_root, exist_ok=True)
        os.makedirs(packed_root, exist_ok=True)

        dataset = SPEECHCOMMANDS(
            speech_commands_root,
            download=True,
            subset="validation",
        )
        indices = torch.randperm(len(dataset))[: len(dataset) // 10].tolist()
        dataset = Subset(dataset, indices)

        class MyTransform(nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.spectrogram_extractor = Spectrogram()

            def forward(self, item):
                waveform = item[0]
                spectrogram = self.spectrogram_extractor(waveform)
                return (spectrogram,) + item[1:]

        transform = MyTransform()
        pack_dataset(
            dataset,
            packed_root,
            transform,
            exists="overwrite",
            num_workers=0,
        )

        # Read from pickle
        from torchoutil.utils.pack import PackedDataset

        packed_root = self.tmpdir.joinpath("packed_speech_commands")
        pack = PackedDataset(packed_root)
        pack[0]  # first transformed item

        # Tests
        indices = torch.randperm(len(dataset))[:10].tolist()

        assert len(dataset) == len(pack)

        for idx in indices:
            item_1 = transform(dataset[idx])
            item_2 = pack[idx]

            assert isinstance(item_1, tuple)
            assert isinstance(item_2, tuple)
            assert len(item_1) == len(item_2)
            assert torch.equal(item_1[0], item_2[0])

            for i in range(1, len(item_1)):
                assert item_1[i] == item_2[i]

        num_steps = 5
        durations_1 = []
        durations_2 = []

        for i in range(num_steps):
            start_1 = time.perf_counter()
            for idx in indices:
                transform(dataset[idx])
            end_1 = time.perf_counter()

            start_2 = time.perf_counter()
            for idx in indices:
                pack[idx]
            end_2 = time.perf_counter()

            durations_1.append(end_1 - start_1)
            durations_2.append(end_2 - start_2)

        duration_1 = sum(durations_1)
        duration_2 = sum(durations_2)
        assert duration_1 > duration_2, f"Found {duration_1} <= {duration_2}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.mask import (
    generate_square_subsequent_mask,
    lengths_to_non_pad_mask,
    lengths_to_pad_mask,
    non_pad_mask_to_lengths,
    pad_mask_to_lengths,
    tensor_to_non_pad_mask,
    tensor_to_pad_mask,
)


class TestTensorToPadMask(TestCase):
    def test_tensor_to_non_pad_mask_example_1(self) -> None:
        inp = torch.as_tensor([1, 10, 20, 2, 0, 0])
        # default include_end value is False
        out = tensor_to_non_pad_mask(inp, end_value=2)
        out_expected = torch.as_tensor([True, True, True, False, False, False])

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_non_pad_mask_example_2(self) -> None:
        inp = torch.as_tensor([1, 10, 20, 2, 0, 0])
        out = tensor_to_non_pad_mask(inp, end_value=2, include_end=True)
        out_expected = torch.as_tensor([True, True, True, True, False, False])

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_non_pad_mask_example_3(self) -> None:
        inp = torch.as_tensor(
            [
                [10, 20, 2, 0, 2],
                [10, 20, 30, 0, 0],
                [2, 2, 0, 99, 10],
            ]
        )
        out = tensor_to_non_pad_mask(inp, end_value=2, include_end=False)
        out_expected = torch.as_tensor(
            [
                [True, True, False, False, False],
                [True, True, True, True, True],
                [False, False, False, False, False],
            ]
        )

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_pad_mask_example_1(self) -> None:
        inp = torch.as_tensor([1, 10, 20, 2, 0, 0])
        out_expected = torch.as_tensor([False, False, False, True, True, True])
        out = tensor_to_pad_mask(inp, end_value=2)  # default include_end value is True

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_pad_mask_example_2(self) -> None:
        inp = torch.as_tensor([1, 10, 20, 2, 0, 0])
        out_expected = torch.as_tensor([False, False, False, False, True, True])
        out = tensor_to_pad_mask(inp, end_value=2, include_end=False)

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_pad_mask_example_3(self) -> None:
        inp = torch.as_tensor([[10, 20, 30, 40, 50], [10, 2, 20, 30, 40]])
        out_expected = torch.as_tensor(
            [[False, False, False, False, False], [False, False, True, True, True]]
        )
        out = tensor_to_pad_mask(inp, end_value=2, include_end=False)

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)

    def test_tensor_to_pad_mask_example_4(self) -> None:
        inp = torch.as_tensor([[10, 20, 30, 40, 50], [10, 2, 20, 30, 40]])
        out_expected = torch.as_tensor(
            [[False, False, False, False, False], [False, True, True, True, True]]
        )
        out = tensor_to_pad_mask(inp, end_value=2, include_end=True)

        assert out.shape == out_expected.shape
        assert torch.equal(out, out_expected)


class TestMaskLengths(TestCase):
    def test_mask_to_lens_to_mask(self) -> None:
        non_pad_mask = torch.as_tensor(
            [[1, 1, 0], [0, 0, 0], [1, 1, 1]], dtype=torch.bool
        )

        out = non_pad_mask_to_lengths(non_pad_mask)
        out_expected = torch.as_tensor([2, 0, 3])
        assert torch.equal(out, out_expected)

        inp2 = out
        out2 = lengths_to_non_pad_mask(inp2, None)
        out2_expected = non_pad_mask
        assert torch.equal(out2, out2_expected), f"{out2=}\n{out2_expected=}"

        pad_mask = torch.as_tensor(
            [
                [0, 1, 1],
                [0, 0, 0],
                [1, 1, 1],
                [0, 0, 1],
            ],
            dtype=torch.bool,
        )
        lens = pad_mask_to_lengths(pad_mask)
        out3 = lengths_to_pad_mask(lens, None)
        out3_expected = pad_mask

        assert torch.equal(out3, out3_expected), f"{out3=}; {out3_expected=}"

    def test_lens_to_non_pad_masks(self) -> None:
        lens = torch.arange(5)

        # Default behaviour
        include_end = False
        expected_mask = torch.as_tensor(
            [
                [0, 0, 0, 0],
                [1, 0, 0, 0],
                [1, 1, 0, 0],
                [1, 1, 1, 0],
                [1, 1, 1, 1],
            ],
            dtype=torch.int,
        )

        non_pad_mask = lengths_to_non_pad_mask(lens, 4).int()
        assert non_pad_mask.shape == expected_mask.shape
        self.assertTrue(
            torch.equal(non_pad_mask, expected_mask),
            f"{include_end=}; {non_pad_mask=}; {expected_mask=}",
        )

        # Special behaviour
        include_end = True
        expected_mask = torch.as_tensor(
            [
                [1, 0, 0, 0],
                [1, 1, 0, 0],
                [1, 1, 1, 0],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
            ],
            dtype=torch.int,
        )

        non_pad_mask = lengths_to_non_pad_mask(lens, 4, include_end).int()
        assert non_pad_mask.shape == expected_mask.shape
        self.assertTrue(
            torch.equal(non_pad_mask, expected_mask),
            f"{include_end=}; {non_pad_mask=}; {expected_mask=}",
        )

    def test_lengths_to_non_pad_mask_example_1(self) -> None:
        input = torch.as_tensor([4, 2, 0, 3, 0])
        output = lengths_to_non_pad_mask(input, max_len=6, include_end=False)
        expected = torch.as_tensor(
            [
                [True, True, True, True, False, False],
                [True, True, False, False, False, False],
                [False, False, False, False, False, False],
                [True, True, True, False, False, False],
                [False, False, False, False, False, False],
            ]
        )

        assert output.shape == expected.shape
        assert torch.equal(output, expected)

    def test_lengths_to_pad_mask_example_1(self) -> None:
        input = torch.as_tensor([4, 2, 0, 3, 0])
        output = lengths_to_pad_mask(input, max_len=None, include_end=True)
        expected = torch.as_tensor(
            [
                [False, False, False, False],
                [False, False, True, True],
                [True, True, True, True],
                [False, False, False, True],
                [True, True, True, True],
            ]
        )

        assert output.shape == expected.shape
        assert torch.equal(output, expected)

    def test_lens_to_pad_masks(self) -> None:
        lens = torch.arange(5)

        # Default behaviour
        include_end = True
        expected_mask = torch.as_tensor(
            [
                [1, 1, 1, 1],
                [0, 1, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 0, 1],
                [0, 0, 0, 0],
            ],
            dtype=torch.int,
        )

        pad_mask = lengths_to_pad_mask(lens, 4).int()
        assert pad_mask.shape == expected_mask.shape
        self.assertTrue(
            torch.equal(pad_mask, expected_mask),
            f"{include_end=}; {pad_mask=}; {expected_mask=}",
        )

        # Special behaviour
        include_end = False
        expected_mask = torch.as_tensor(
            [
                [0, 1, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 0, 1],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ],
            dtype=torch.int,
        )

        pad_mask = lengths_to_pad_mask(lens, 4, include_end).int()
        assert pad_mask.shape == expected_mask.shape
        self.assertTrue(
            torch.equal(pad_mask, expected_mask),
            f"{include_end=}; {pad_mask=}; {expected_mask=}",
        )


class TestGenerateSqMask(TestCase):
    def test_generate_square_subsequent_mask_example_1(self) -> None:
        inf = math.inf
        output = generate_square_subsequent_mask(6)
        expected = torch.as_tensor(
            [
                [0.0, -inf, -inf, -inf, -inf, -inf],
                [0.0, 0.0, -inf, -inf, -inf, -inf],
                [0.0, 0.0, 0.0, -inf, -inf, -inf],
                [0.0, 0.0, 0.0, 0.0, -inf, -inf],
                [0.0, 0.0, 0.0, 0.0, 0.0, -inf],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        assert output.shape == expected.shape
        assert torch.equal(output, expected)

    def test_generate_shifted_sq_mask_example_1(self) -> None:
        inf = math.inf
        output = generate_square_subsequent_mask(6, diagonal=2)
        expected = torch.as_tensor(
            [
                [0.0, 0.0, 0.0, -inf, -inf, -inf],
                [0.0, 0.0, 0.0, 0.0, -inf, -inf],
                [0.0, 0.0, 0.0, 0.0, 0.0, -inf],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        assert output.shape == expected.shape
        assert torch.equal(output, expected)

    def test_generate_shifted_sq_mask_example_2(self) -> None:
        inf = math.inf
        output = generate_square_subsequent_mask(6, diagonal=-2)
        expected = torch.as_tensor(
            [
                [-inf, -inf, -inf, -inf, -inf, -inf],
                [-inf, -inf, -inf, -inf, -inf, -inf],
                [0.0, -inf, -inf, -inf, -inf, -inf],
                [0.0, 0.0, -inf, -inf, -inf, -inf],
                [0.0, 0.0, 0.0, -inf, -inf, -inf],
                [0.0, 0.0, 0.0, 0.0, -inf, -inf],
            ]
        )
        assert output.shape == expected.shape
        assert torch.equal(output, expected)

    def test_generate_shifted_sq_mask_example_3(self) -> None:
        inf = math.inf
        output = generate_square_subsequent_mask(6, diagonal=4)
        expected = torch.as_tensor(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, -inf],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        assert output.shape == expected.shape
        assert torch.equal(output, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.segments import extract_segments


class TestSegments(TestCase):
    def test_example_1(self) -> None:
        x = torch.as_tensor([0, 1, 1, 0, 0, 1, 1, 1, 1, 0]).bool()
        starts, ends = extract_segments(x)

        assert torch.equal(starts, torch.as_tensor([1, 5]))
        assert torch.equal(ends, torch.as_tensor([3, 9]))

    def test_example_2(self) -> None:
        x = torch.as_tensor([[1, 1, 1, 0], [1, 0, 0, 1]]).bool()
        indices, starts, ends = extract_segments(x)

        assert torch.equal(indices, torch.as_tensor([0, 1, 1]))
        assert torch.equal(starts, torch.as_tensor([0, 0, 3]))
        assert torch.equal(ends, torch.as_tensor([3, 1, 4]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from dataclasses import dataclass
from typing import NamedTuple
from unittest import TestCase

from torchoutil.pyoutil import (
    is_dataclass_instance,
    is_iterable_str,
    is_namedtuple_instance,
)


class TestTypeChecks(TestCase):
    def test_is_iterable_str_1(self) -> None:
        inputs = [
            ("a", True, False),
            (["a"], True, True),
            ([], True, True),
            (("a",), True, True),
            ((), True, True),
            (1.0, False, False),
            ([["a"]], False, False),
            ("", True, False),
            ((s for s in ("a", "b", "c")), True, True),
        ]

        for x, expected_1, expected_2 in inputs:
            result_1 = is_iterable_str(x, accept_str=True)
            result_2 = is_iterable_str(x, accept_str=False)

            assert expected_1 == result_1
            assert expected_2 == result_2

    def test_is_dataclass_example_1(self) -> None:
        @dataclass
        class DC:
            a: int = 0
            b: str = "0"

        dc = DC(a=0, b="0")

        assert not is_namedtuple_instance(DC)
        assert not is_namedtuple_instance(dc)

        assert not is_dataclass_instance(DC)
        assert is_dataclass_instance(dc)

    def test_is_namedtuple_example_1(self) -> None:
        class NT1(NamedTuple):
            a: int
            b: str

        NT2 = NamedTuple("NT2", [("a", int), ("b", str)])

        nt1 = NT1(a=0, b="0")
        nt2 = NT2(a=0, b="0")

        assert not is_namedtuple_instance(NT1)
        assert not is_namedtuple_instance(NT2)
        assert is_namedtuple_instance(nt1)
        assert is_namedtuple_instance(nt2)

        assert not is_dataclass_instance(NT1)
        assert not is_dataclass_instance(NT2)
        assert not is_dataclass_instance(nt1)
        assert not is_dataclass_instance(nt2)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
import logging
import sys
from functools import lru_cache
from logging import FileHandler, Formatter, Logger, StreamHandler
from pathlib import Path
from types import ModuleType
from typing import IO, List, Literal, Optional, Sequence, TypeVar, Union

from .importlib import reload_submodules

T = TypeVar("T", covariant=True)

PackageOrLogger = Union[str, ModuleType, None, Logger, Literal["__parent_file__"]]
PackageOrLoggerList = Union[PackageOrLogger, Sequence[PackageOrLogger]]

_PARENT_FILE_KEY = "__parent_file__"
DEFAULT_FMT = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
VERBOSE_DEBUG = 2
VERBOSE_INFO = 1
VERBOSE_WARNING = 0
VERBOSE_ERROR = -1

pylog = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def warn_once(
    msg: str,
    logger: PackageOrLoggerList = _PARENT_FILE_KEY,
    *,
    level: int = logging.WARNING,
) -> None:
    loggers = _get_loggers(logger)
    for logger in loggers:
        logger.log(level, msg)


def setup_logging_verbose(
    package_or_logger: PackageOrLoggerList = None,
    verbose: Optional[int] = VERBOSE_INFO,
    *,
    fmt: Union[str, None, Formatter] = DEFAULT_FMT,
    stream: Union[IO[str], Literal["auto"]] = "auto",
    set_fmt: bool = True,
    capture_warnings: bool = True,
    autoreload: bool = True,
) -> None:
    """Helper function to customize logging messages using verbose_level.

    Note: Higher verbose values means more debug messages.
    """
    if verbose is None:
        level = None
    else:
        level = _verbose_to_logging_level(verbose)

    return setup_logging_level(
        package_or_logger,
        level=level,
        fmt=fmt,
        stream=stream,
        set_fmt=set_fmt,
        capture_warnings=capture_warnings,
        autoreload=autoreload,
    )


def setup_logging_level(
    package_or_logger: PackageOrLoggerList = None,
    level: Optional[int] = logging.INFO,
    *,
    fmt: Union[str, None, Formatter] = DEFAULT_FMT,
    stream: Union[IO[str], Literal["auto"]] = "auto",
    set_fmt: bool = True,
    capture_warnings: bool = True,
    autoreload: bool = True,
) -> None:
    """Helper function to customize logging messages using logging.level.

    Note: Lower level values means more debug messages.
    """
    logging.captureWarnings(capture_warnings)

    logger_lst = _get_loggers(package_or_logger)
    if isinstance(fmt, str):
        fmt = Formatter(fmt)

    if stream == "auto":
        if running_on_interpreter():
            stream = sys.stdout
        else:
            stream = sys.stderr

    for logger in logger_lst:
        if set_fmt:
            found = False

            for handler in logger.handlers:
                if isinstance(handler, StreamHandler) and handler.stream is stream:
                    handler.setFormatter(fmt)
                    found = True
                    break

            if not found:
                handler = StreamHandler(stream)  # type: ignore
                handler.setFormatter(fmt)
                logger.addHandler(handler)

        if level is not None:
            logger.setLevel(level)

    if autoreload:
        for logger in logger_lst:
            if logger.name not in sys.modules:
                continue
            reload_submodules(sys.modules[logger.name])


def running_on_interpreter() -> bool:
    return get_ipython_name() is None


def running_on_terminal() -> bool:
    return get_ipython_name() == "TerminalInteractiveShell"


def running_on_notebook() -> bool:
    return get_ipython_name() == "ZMQInteractiveShell"


def get_ipython_name() -> (
    Optional[Literal["TerminalInteractiveShell", "ZMQInteractiveShell"]]
):
    try:
        return get_ipython().__class__.__name__  # type: ignore
    except NameError:
        return None


def get_current_file_logger(
    *,
    parent_deep: int = 1,
    default: T = logging.root,
) -> Union[Logger, T]:
    """Returns the logger of the caller file. If this cannot be found, returns the root logger."""
    try:
        frame = inspect.currentframe()
        for _ in range(parent_deep):
            frame = frame.f_back  # type: ignore
        parent_name = frame.f_globals["__name__"]  # type: ignore
        return logging.getLogger(parent_name)
    except (AttributeError, KeyError):
        return default


def get_null_logger() -> Logger:
    logger = logging.getLogger("null_logger")
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.CRITICAL + 1)
    return logger


def _get_loggers(pkg_name_log_arg: PackageOrLoggerList) -> List[Logger]:
    if pkg_name_log_arg is None or isinstance(
        pkg_name_log_arg, (str, Logger, ModuleType)
    ):
        pkg_name_log_lst = [pkg_name_log_arg]
    else:
        pkg_name_log_lst = list(pkg_name_log_arg)

    loggers: List[Logger] = []
    for pkg_name_log in pkg_name_log_lst:
        if isinstance(pkg_name_log, ModuleType):
            logger = logging.getLogger(pkg_name_log.__name__)

        elif pkg_name_log == _PARENT_FILE_KEY:
            logger = get_current_file_logger(parent_deep=2)

        elif isinstance(pkg_name_log, (type(None), str)):
            logger = logging.getLogger(pkg_name_log)

        else:
            logger = pkg_name_log

        loggers.append(logger)

    return loggers


class MkdirFileHandler(FileHandler):
    """FileHandler that build intermediate directories to filename.

    Used for export hydra logs to a file contained in a folder that does not exists yet at the start of the program.
    """

    def __init__(
        self,
        filename: Union[str, Path],
        mode: str = "a",
        encoding: Optional[str] = None,
        delay: bool = True,
        errors: Optional[str] = None,
        mkdir_parents: bool = True,
        mkdir_exist_ok: bool = True,
    ) -> None:
        filename = Path(filename)
        filename.parent.mkdir(parents=mkdir_parents, exist_ok=mkdir_exist_ok)

        super().__init__(filename, mode, encoding, delay, errors)


def _verbose_to_logging_level(verbose: int) -> int:
    if verbose <= VERBOSE_ERROR:
        level = logging.ERROR
    elif verbose == VERBOSE_WARNING:
        level = logging.WARNING
    elif verbose == VERBOSE_INFO:
        level = logging.INFO
    else:
        level = logging.DEBUG
    return level



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchoutil.nn.functional.powerset import (
    multilabel_to_powerset,
    powerset_to_multilabel,
)


class TestPowerset(TestCase):
    def test_example_1(self) -> None:
        multilabel = torch.as_tensor(
            [[0, 0, 0, 1], [1, 0, 1, 0], [0, 1, 1, 0], [1, 0, 0, 0]]
        )[None].float()
        num_classes = multilabel.shape[-1]
        max_set_size = int(multilabel.view(-1, num_classes).sum(-1).max().int().item())

        powerset = multilabel_to_powerset(
            multilabel,
            num_classes=num_classes,
            max_set_size=max_set_size,
        )
        multilabel_back = powerset_to_multilabel(
            powerset,
            num_classes=num_classes,
            max_set_size=max_set_size,
        )
        assert torch.equal(multilabel, multilabel_back)


if __name__ == "__main__":
    unittest.main()
